import decimal
from _typeshed import Incomplete
from typing import Callable, Final

ABI_DECIMAL_PREC: Final[int]
abi_decimal_context: Final[Incomplete]
decimal_localcontext: Final[Incomplete]
ZERO: Final[Incomplete]
TEN: Final[Incomplete]
Decimal: Final[Incomplete]

def ceil32(x: int) -> int: ...
def compute_unsigned_integer_bounds(num_bits: int) -> tuple[int, int]: ...
def compute_signed_integer_bounds(num_bits: int) -> tuple[int, int]: ...
def compute_unsigned_fixed_bounds(
    num_bits: int, frac_places: int
) -> tuple[decimal.Decimal, decimal.Decimal]: ...
def compute_signed_fixed_bounds(
    num_bits: int, frac_places: int
) -> tuple[decimal.Decimal, decimal.Decimal]: ...
def scale_places(places: int) -> Callable[[decimal.Decimal], decimal.Decimal]:
    """
    Returns a function that shifts the decimal point of decimal values to the
    right by ``places`` places.
    """



import pytest

from faster_eth_abi.utils.numeric import (
    ceil32,
)


@pytest.mark.parametrize(
    "value,expected",
    (
        (0, 0),
        (1, 32),
        (31, 32),
        (32, 32),
        (33, 64),
        (63, 64),
        (64, 64),
        (100, 128),
    ),
)
def test_ceil32(value, expected):
    actual = ceil32(value)
    assert actual == expected



import parsimonious
from _typeshed import Incomplete
from faster_eth_abi._grammar import (
    ABIType as ABIType,
    BasicType as BasicType,
    TYPE_ALIASES as TYPE_ALIASES,
    TYPE_ALIAS_RE as TYPE_ALIAS_RE,
    TupleType as TupleType,
    normalize as normalize,
)
from typing import Final

__all__ = [
    "NodeVisitor",
    "ABIType",
    "TupleType",
    "BasicType",
    "grammar",
    "parse",
    "normalize",
    "visitor",
    "TYPE_ALIASES",
    "TYPE_ALIAS_RE",
]

grammar: Final[Incomplete]

class NodeVisitor(parsimonious.NodeVisitor):
    """
    Parsimonious node visitor which performs both parsing of type strings and
    post-processing of parse trees.  Parsing operations are cached.
    """

    parse: Incomplete
    def __init__(self) -> None: ...
    grammar = grammar
    def visit_non_zero_tuple(self, node, visited_children): ...
    def visit_tuple_type(self, node, visited_children): ...
    def visit_next_type(self, node, visited_children): ...
    def visit_basic_type(self, node, visited_children): ...
    def visit_two_size(self, node, visited_children): ...
    def visit_const_arr(self, node, visited_children): ...
    def visit_dynam_arr(self, node, visited_children): ...
    def visit_alphas(self, node, visited_children): ...
    def visit_digits(self, node, visited_children): ...
    def generic_visit(self, node, visited_children): ...

visitor: Final[Incomplete]
parse: Final[Incomplete]



import re

from hypothesis import (
    event,
    given,
    settings,
    strategies as st,
)

from faster_eth_abi.registry import (
    BaseEquals,
    Equals,
    has_arrlist,
    is_base_tuple,
)
from tests.core.common.strategies import (
    malformed_type_strs,
    type_strs,
)


@given(st.text())
def test_equals_has_expected_behavior(s):
    pos_predicate = Equals(s)
    neg_predicate = Equals(s + "a")

    # Matches correct strings
    assert pos_predicate(s)
    assert not neg_predicate(s)

    # Can be equal/unequal
    assert pos_predicate == Equals(s)
    assert pos_predicate != neg_predicate


def test_equals_can_be_dict_key():
    pred = Equals("foo")
    assert {pred: "bar"}[pred] == "bar"


def test_equals_has_expected_repr():
    assert repr(Equals("foo")) == "<Equals (== 'foo')>"


ARRAY_RE = re.compile(r"\[\d*\]")


@settings(max_examples=200)
@given(type_strs)
def test_base_equals_has_expected_behavior_for_parsable_types(type_str):
    is_int = BaseEquals("int")
    is_int_with_sub = BaseEquals("int", with_sub=True)
    is_int_with_no_sub = BaseEquals("int", with_sub=False)

    # Should not match tuple types
    if type_str.startswith("("):
        assert not is_int(type_str)
        assert not is_int_with_sub(type_str)
        assert not is_int_with_no_sub(type_str)
        event("No match for tuple type")

    # Should not match array types
    elif ARRAY_RE.search(type_str):
        assert not is_int(type_str)
        assert not is_int_with_sub(type_str)
        assert not is_int_with_no_sub(type_str)
        event("No match for array type")

    # Should match types with int base
    elif type_str.startswith("int"):
        assert is_int(type_str)
        event("Match for base")

        if type_str == "int":
            assert is_int_with_no_sub(type_str)
            assert not is_int_with_sub(type_str)
            event("Match for base with no sub")
        else:
            assert not is_int_with_no_sub(type_str)
            assert is_int_with_sub(type_str)
            event("Match for base with sub")

    # Should not match any other types
    else:
        assert not is_int(type_str)
        assert not is_int_with_sub(type_str)
        assert not is_int_with_no_sub(type_str)
        event("No match for other base")


def test_base_equals_can_be_dict_key():
    pred = BaseEquals("foo")
    assert {pred: "bar"}[pred] == "bar"


def test_base_equals_has_expected_repr():
    assert repr(BaseEquals("foo")) == "<BaseEquals (base == 'foo')>"
    assert (
        repr(BaseEquals("foo", with_sub=True))
        == "<BaseEquals (base == 'foo' and sub is not None)>"
    )
    assert (
        repr(BaseEquals("foo", with_sub=False))
        == "<BaseEquals (base == 'foo' and sub is None)>"
    )


@given(type_strs)
def test_has_arrlist_has_expected_behavior_for_parsable_types(type_str):
    # Should not match tuple types
    if type_str.startswith("("):
        assert not has_arrlist(type_str)
        event("No match for tuple type")

    # Should match array types
    elif ARRAY_RE.search(type_str):
        assert has_arrlist(type_str)
        event("Match for array type")

    # Should not match any other types
    else:
        assert not has_arrlist(type_str)
        event("No match for non-array type")


@given(type_strs)
def test_is_base_tuple_has_expected_behavior_for_parsable_types(type_str):
    # Should match tuple types
    if type_str.endswith(")"):
        assert is_base_tuple(type_str)
        event("Match for tuple type")

    # Should not match any other types
    else:
        assert not is_base_tuple(type_str)
        event("No match for non-tuple type")


@given(malformed_type_strs)
def test_predicates_have_expected_behavior_for_malformed_types(malformed_type_str):
    is_int = BaseEquals("int")
    is_int_with_sub = BaseEquals("int", with_sub=True)
    is_int_with_no_sub = BaseEquals("int", with_sub=False)

    # Should not match unparsable types
    assert not is_int(malformed_type_str)
    assert not is_int_with_sub(malformed_type_str)
    assert not is_int_with_no_sub(malformed_type_str)

    assert not has_arrlist(malformed_type_str)
    assert not is_base_tuple(malformed_type_str)



from faster_eth_abi.encoding import (
    BaseEncoder as BaseEncoder,
    TupleEncoder as TupleEncoder,
)
from faster_eth_abi.exceptions import ValueOutOfBounds as ValueOutOfBounds
from typing import Any, Callable, Sequence, TypeVar

T = TypeVar("T")

def validate_tuple(self, value: Sequence[Any]) -> None: ...
def encode_tuple(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_all_dynamic(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic1(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic2(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic3(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic4(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic5(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic6(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic7(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic8(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic9(self, values: Sequence[Any]) -> bytes: ...
def encode_tuple_no_dynamic10(self, values: Sequence[Any]) -> bytes: ...

encode_tuple_no_dynamic_funcs: dict[int, Callable[[TupleEncoder, Sequence[Any]], bytes]]

def encode_fixed(
    value: Any,
    encode_fn: Callable[[Any], bytes],
    is_big_endian: bool,
    data_byte_size: int,
) -> bytes: ...
def encode_signed(
    value: T, encode_fn: Callable[[T], bytes], data_byte_size: int
) -> bytes: ...
def encode_elements(item_encoder: BaseEncoder, value: Sequence[Any]) -> bytes: ...
def encode_elements_dynamic(
    item_encoder: BaseEncoder, value: Sequence[Any]
) -> bytes: ...
def encode_uint_256(i: int) -> bytes: ...
def int_to_big_endian(value: int) -> bytes: ...



import pytest

from faster_eth_abi.grammar import (
    parse,
)
from faster_eth_abi.packed import (
    encode_packed,
)
from tests.core.common.unit import (
    CORRECT_ENCODINGS,
    CORRECT_TUPLE_ENCODINGS,
)


@pytest.mark.parametrize(
    "single_abi_type,python_value,_,packed_encoding",
    CORRECT_TUPLE_ENCODINGS,
)
def test_encode_packed(single_abi_type, python_value, _, packed_encoding):
    abi_type = parse(single_abi_type)
    if abi_type.arrlist is not None:
        pytest.skip("ABI coding functions do not support array types")

    types = [t.to_type_str() for t in abi_type.components]

    actual = encode_packed(types, python_value)
    assert actual == packed_encoding


@pytest.mark.parametrize(
    "single_abi_type,python_value,_,packed_encoding",
    CORRECT_ENCODINGS,
)
def test_encode_packed_single_types(single_abi_type, python_value, _, packed_encoding):
    actual = encode_packed([single_abi_type], [python_value])
    assert actual == packed_encoding



"""Private helpers for decoding logic, intended for C compilation.

This file exists because the original decoding.py is not ready to be fully compiled to C.
This module contains functions and logic that we wish to compile.
"""
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Final,
    Tuple,
)

from faster_eth_utils import (
    big_endian_to_int,
)

from faster_eth_abi.exceptions import (
    InsufficientDataBytes,
    InvalidPointer,
    NonEmptyPaddingBytes,
)
from faster_eth_abi.io import (
    BytesIO,
    ContextFramesBytesIO,
)
from faster_eth_abi.typing import (
    T,
)

if TYPE_CHECKING:
    from .decoding import (
        BaseArrayDecoder,
        DynamicArrayDecoder,
        FixedByteSizeDecoder,
        HeadTailDecoder,
        SignedIntegerDecoder,
        SizedArrayDecoder,
        TupleDecoder,
    )


# Helpers
def decode_uint_256(stream: ContextFramesBytesIO) -> int:
    """
    A faster version of :func:`~decoding.decode_uint_256` in decoding.py.

    It recreates the logic from the UnsignedIntegerDecoder, but we can
    skip a lot because we know the value of many vars.
    """
    # read data from stream
    if len(data := stream.read(32)) == 32:
        return big_endian_to_int(data)  # type: ignore [no-any-return]
    raise InsufficientDataBytes(f"Tried to read 32 bytes, only got {len(data)} bytes.")


def get_value_byte_size(decoder: "FixedByteSizeDecoder") -> int:
    return decoder.value_bit_size // 8


# HeadTailDecoder
def decode_head_tail(self: "HeadTailDecoder[T]", stream: ContextFramesBytesIO) -> T:
    # Decode the offset and move the stream cursor forward 32 bytes
    start_pos = decode_uint_256(stream)
    # Jump ahead to the start of the value
    stream.push_frame(start_pos)

    # assertion check for mypy
    tail_decoder = self.tail_decoder
    if tail_decoder is None:
        raise AssertionError("`tail_decoder` is None")
    # Decode the value
    value: T = tail_decoder(stream)
    # Return the cursor
    stream.pop_frame()

    return value


# TupleDecoder
def decode_tuple(
    self: "TupleDecoder[T]", stream: ContextFramesBytesIO
) -> Tuple[T, ...]:
    # NOTE: the original implementation would do this but it's
    # kinda wasteful, so we rebuilt the logic within this function
    # validate_pointers_tuple(self, stream)

    current_location = stream.tell()
    if self._no_head_tail:
        # TODO: if all(isinstance(d, TupleDecoder) for d in self._decoders)
        #           return tuple(decode_tuple(stream) for _ in range(len(self.decoders))
        #       and other types with compiled decode funcs
        return tuple(decoder(stream) for decoder in self.decoders)

    end_of_offsets = current_location + 32 * self.len_of_head
    total_stream_length = len(stream.getbuffer())
    items = []
    for decoder, is_head_tail in zip(self.decoders, self._is_head_tail):
        if is_head_tail:
            # the next 32 bytes are a pointer that we should validate
            # checkpoint the stream location so we can reset it after validation
            step_location = stream.tell()

            offset = decode_uint_256(stream)
            indicated_idx = current_location + offset
            if indicated_idx < end_of_offsets or indicated_idx >= total_stream_length:
                # the pointer is indicating its data is located either within the
                # offsets section of the stream or beyond the end of the stream,
                # both of which are invalid
                raise InvalidPointer(
                    "Invalid pointer in tuple at location "
                    f"{stream.tell() - 32} in payload"
                )

            # reset the stream so we can decode
            stream.seek(step_location)

        items.append(decoder(stream))

    # return the stream to its original location for actual decoding
    stream.seek(current_location)

    return tuple(items)


def validate_pointers_tuple(
    self: "TupleDecoder",
    stream: ContextFramesBytesIO,
) -> None:
    """
    Verify that all pointers point to a valid location in the stream.
    """
    current_location = stream.tell()
    if self._no_head_tail:
        for decoder in self.decoders:
            decoder(stream)
    else:
        end_of_offsets = current_location + 32 * self.len_of_head
        total_stream_length = len(stream.getbuffer())
        for decoder, is_head_tail in zip(self.decoders, self._is_head_tail):
            if not is_head_tail:
                # the next 32 bytes are not a pointer,
                # so progress the stream per the decoder
                decoder(stream)
            else:
                # the next 32 bytes are a pointer
                offset = decode_uint_256(stream)
                indicated_idx = current_location + offset
                if (
                    indicated_idx < end_of_offsets
                    or indicated_idx >= total_stream_length
                ):
                    # the pointer is indicating its data is located either within the
                    # offsets section of the stream or beyond the end of the stream,
                    # both of which are invalid
                    raise InvalidPointer(
                        "Invalid pointer in tuple at location "
                        f"{stream.tell() - 32} in payload"
                    )
    # return the stream to its original location for actual decoding
    stream.seek(current_location)


# BaseArrayDecoder
def validate_pointers_array(
    self: "BaseArrayDecoder", stream: ContextFramesBytesIO, array_size: int
) -> None:
    """
    Verify that all pointers point to a valid location in the stream.
    """
    current_location = stream.tell()
    end_of_offsets = current_location + 32 * array_size
    total_stream_length = len(stream.getbuffer())
    for _ in range(array_size):
        offset = decode_uint_256(stream)
        indicated_idx = current_location + offset
        if indicated_idx < end_of_offsets or indicated_idx >= total_stream_length:
            # the pointer is indicating its data is located either within the
            # offsets section of the stream or beyond the end of the stream,
            # both of which are invalid
            raise InvalidPointer(
                "Invalid pointer in array at location "
                f"{stream.tell() - 32} in payload"
            )
    stream.seek(current_location)


# SizedArrayDecoder
def decode_sized_array(
    self: "SizedArrayDecoder[T]", stream: ContextFramesBytesIO
) -> Tuple[T, ...]:
    item_decoder = self.item_decoder
    if item_decoder is None:
        raise AssertionError("`item_decoder` is None")

    array_size = self.array_size
    self.validate_pointers(stream, array_size)
    return tuple(item_decoder(stream) for _ in range(array_size))


# DynamicArrayDecoder
def decode_dynamic_array(
    self: "DynamicArrayDecoder[T]", stream: ContextFramesBytesIO
) -> Tuple[T, ...]:
    array_size = decode_uint_256(stream)
    stream.push_frame(32)
    if self.item_decoder is None:
        raise AssertionError("`item_decoder` is None")

    self.validate_pointers(stream, array_size)
    item_decoder = self.item_decoder
    try:
        return tuple(item_decoder(stream) for _ in range(array_size))
    finally:
        stream.pop_frame()


# FixedByteSizeDecoder
def read_fixed_byte_size_data_from_stream(
    self: "FixedByteSizeDecoder[Any]",
    # NOTE: use BytesIO here so mypyc doesn't type-check
    # `stream` once we compile ContextFramesBytesIO.
    stream: BytesIO,
) -> bytes:
    data_byte_size = self.data_byte_size
    if len(data := stream.read(data_byte_size)) == data_byte_size:
        return data
    raise InsufficientDataBytes(
        f"Tried to read {data_byte_size} bytes, only got {len(data)} bytes."
    )


def split_data_and_padding_fixed_byte_size(
    self: "FixedByteSizeDecoder[Any]",
    raw_data: bytes,
) -> Tuple[bytes, bytes]:
    value_byte_size = get_value_byte_size(self)
    padding_size = self.data_byte_size - value_byte_size

    if self.is_big_endian:
        if padding_size == 0:
            return raw_data, b""
        padding_bytes = raw_data[:padding_size]
        data = raw_data[padding_size:]
    else:
        data = raw_data[:value_byte_size]
        padding_bytes = raw_data[value_byte_size:]

    return data, padding_bytes


def validate_padding_bytes_fixed_byte_size(
    self: "FixedByteSizeDecoder[T]",
    value: T,
    padding_bytes: bytes,
) -> None:
    if padding_bytes != get_expected_padding_bytes(self, b"\x00"):
        raise NonEmptyPaddingBytes(f"Padding bytes were not empty: {padding_bytes!r}")


_expected_padding_bytes_cache: Final[
    Dict["FixedByteSizeDecoder[Any]", Dict[bytes, bytes]]
] = {}


def get_expected_padding_bytes(
    self: "FixedByteSizeDecoder[Any]", chunk: bytes
) -> bytes:
    instance_cache = _expected_padding_bytes_cache.setdefault(self, {})
    expected_padding_bytes = instance_cache.get(chunk)
    if expected_padding_bytes is None:
        value_byte_size = get_value_byte_size(self)
        padding_size = self.data_byte_size - value_byte_size
        expected_padding_bytes = chunk * padding_size
        instance_cache[chunk] = expected_padding_bytes
    return expected_padding_bytes


def validate_padding_bytes_signed_integer(
    self: "SignedIntegerDecoder",
    value: int,
    padding_bytes: bytes,
) -> None:
    if value >= 0:
        expected_padding_bytes = get_expected_padding_bytes(self, b"\x00")
    else:
        expected_padding_bytes = get_expected_padding_bytes(self, b"\xff")

    if padding_bytes != expected_padding_bytes:
        raise NonEmptyPaddingBytes(f"Padding bytes were not empty: {padding_bytes!r}")


# BooleanDecoder
def decoder_fn_boolean(data: bytes) -> bool:
    if data == b"\x00":
        return False
    elif data == b"\x01":
        return True
    raise NonEmptyPaddingBytes(f"Boolean must be either 0x0 or 0x1.  Got: {data!r}")



"""Context-aware byte stream for ABI decoding.

Implements a BytesIO subclass that supports contextual frame management for nested ABI decoding.
"""
from io import (
    BytesIO,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    List,
    Tuple,
    final,
)

from mypy_extensions import (
    mypyc_attr,
)

if TYPE_CHECKING:
    from _typeshed import (
        ReadableBuffer,
    )


@final
@mypyc_attr(allow_interpreted_subclasses=True)
class ContextFramesBytesIO(BytesIO):
    """
    A byte stream which can track a series of contextual frames in a stack. This
    data structure is necessary to perform nested decodings using the
    :py:class:``HeadTailDecoder`` since offsets present in head sections are
    relative only to a particular encoded object.  These offsets can only be
    used to locate a position in a decoding stream if they are paired with a
    contextual offset that establishes the position of the object in which they
    are found.

    For example, consider the encoding of a value for the following type::

        type: (int,(int,int[]))
        value: (1,(2,[3,3]))

    There are two tuples in this type: one inner and one outer.  The inner tuple
    type contains a dynamic type ``int[]`` and, therefore, is itself dynamic.
    This means that its value encoding will be placed in the tail section of the
    outer tuple's encoding.  Furthermore, the inner tuple's encoding will,
    itself, contain a tail section with the encoding for ``[3,3]``.  All
    together, the encoded value of ``(1,(2,[3,3]))`` would look like this (the
    data values are normally 32 bytes wide but have been truncated to remove the
    redundant zeros at the beginnings of their encodings)::

                       offset data
        --------------------------
             ^              0 0x01
             |             32 0x40 <-- Offset of object A in global frame (64)
        -----|--------------------
        Global frame ^     64 0x02 <-- Beginning of object A (64 w/offset 0 = 64)
             |       |     96 0x40 <-- Offset of object B in frame of object A (64)
        -----|-Object A's frame---
             |       |    128 0x02 <-- Beginning of object B (64 w/offset 64 = 128)
             |       |    160 0x03
             v       v    192 0x03
        --------------------------

    Note that the offset of object B is encoded as 64 which only specifies the
    beginning of its encoded value relative to the beginning of object A's
    encoding.  Globally, object B is located at offset 128.  In order to make
    sense out of object B's offset, it needs to be positioned in the context of
    its enclosing object's frame (object A).
    """

    def __init__(self, initial_bytes: "ReadableBuffer"):
        super().__init__(initial_bytes)

        self._frames: Final[List[Tuple[int, int]]] = []
        self._total_offset = 0

    def seek_in_frame(self, pos: int, *args: Any, **kwargs: Any) -> None:
        """
        Seeks relative to the total offset of the current contextual frames.
        """
        self.seek(self._total_offset + pos, *args, **kwargs)

    def push_frame(self, offset: int) -> None:
        """
        Pushes a new contextual frame onto the stack with the given offset and a
        return position at the current cursor position then seeks to the new
        total offset.
        """
        self._frames.append((offset, self.tell()))
        self._total_offset += offset

        self.seek_in_frame(0)

    def pop_frame(self):
        """
        Pops the current contextual frame off of the stack and returns the
        cursor to the frame's return position.
        """
        try:
            offset, return_pos = self._frames.pop()
        except IndexError:
            raise IndexError("no frames to pop")
        self._total_offset -= offset

        self.seek(return_pos)



import pytest

from hypothesis import (
    example,
    given,
    settings,
    strategies as st,
)

from faster_eth_abi.exceptions import (
    NoEntriesFound,
    ParseError,
)
from faster_eth_abi.tools import (
    get_abi_strategy,
)
from tests.core.common.strategies import (
    malformed_type_strs,
    type_strs,
)


@given(type_strs)
@settings(deadline=None)
def test_get_abi_strategy_returns_strategy_for_valid_type_strs(type_str):
    assert isinstance(get_abi_strategy(type_str), st.SearchStrategy)


@given(malformed_type_strs)
@example("address20")
def test_get_abi_strategy_returns_no_strategy_for_invalid_type_strs(malformed_type_str):
    try:
        get_abi_strategy(malformed_type_str)
    except (ParseError, NoEntriesFound, ValueError):
        pass
    else:
        raise AssertionError("Expected ParseError or NoEntriesFound")


@pytest.mark.parametrize(
    "type_str,strategy_repr",
    (
        ("uint8", "integers(min_value=0, max_value=255)"),
        ("uint16", "integers(min_value=0, max_value=65535)"),
        (
            "uint",
            "integers(min_value=0, max_value=115792089237316195423570985008687907853269984665640564039457584007913129639935)",  # noqa: E501
        ),
        (
            "uint256",
            "integers(min_value=0, max_value=115792089237316195423570985008687907853269984665640564039457584007913129639935)",  # noqa: E501
        ),
        ("int8", "integers(min_value=-128, max_value=127)"),
        ("int16", "integers(min_value=-32768, max_value=32767)"),
        (
            "int",
            "integers(min_value=-57896044618658097711785492504343953926634992332820282019728792003956564819968, max_value=57896044618658097711785492504343953926634992332820282019728792003956564819967)",  # noqa: E501
        ),
        (
            "int256",
            "integers(min_value=-57896044618658097711785492504343953926634992332820282019728792003956564819968, max_value=57896044618658097711785492504343953926634992332820282019728792003956564819967)",  # noqa: E501
        ),
        ("address", "binary(min_size=20, max_size=20).map(to_checksum_address)"),
        ("bool", "booleans()"),
        (
            "ufixed8x10",
            "decimals(min_value=0, max_value=255, places=0).map(scale_by_Eneg10)",
        ),
        (
            "ufixed16x11",
            "decimals(min_value=0, max_value=65535, places=0).map(scale_by_Eneg11)",
        ),
        (
            "ufixed",
            "decimals(min_value=0, max_value=340282366920938463463374607431768211455, places=0).map(scale_by_Eneg18)",  # noqa: E501
        ),
        (
            "ufixed128x18",
            "decimals(min_value=0, max_value=340282366920938463463374607431768211455, places=0).map(scale_by_Eneg18)",  # noqa: E501
        ),
        (
            "fixed8x8",
            "decimals(min_value=-128, max_value=127, places=0).map(scale_by_Eneg8)",
        ),
        (
            "fixed16x7",
            "decimals(min_value=-32768, max_value=32767, places=0).map(scale_by_Eneg7)",
        ),
        (
            "fixed",
            "decimals(min_value=-170141183460469231731687303715884105728, max_value=170141183460469231731687303715884105727, places=0).map(scale_by_Eneg18)",  # noqa: E501
        ),
        (
            "fixed128x18",
            "decimals(min_value=-170141183460469231731687303715884105728, max_value=170141183460469231731687303715884105727, places=0).map(scale_by_Eneg18)",  # noqa: E501
        ),
        ("bytes2", "binary(min_size=2, max_size=2)"),
        ("bytes", "binary(max_size=4096)"),
        ("function", "binary(min_size=24, max_size=24)"),
        ("string", "text()"),
        ("bool[]", "lists(booleans())"),
        ("bool[3]", "lists(booleans(), min_size=3, max_size=3)"),
        (
            "bool[3][]",
            "lists(lists(booleans(), min_size=3, max_size=3))",
        ),
        ("(bool,int8)", "tuples(booleans(), integers(min_value=-128, max_value=127))"),
        (
            "(bool,(uint8,int8))",
            "tuples(booleans(), tuples(integers(min_value=0, max_value=255), integers(min_value=-128, max_value=127)))",  # noqa: E501
        ),
        ("(bool,bool)[]", "lists(tuples(booleans(), booleans()))"),
    ),
)
def test_get_abi_strategy_returns_certain_strategies_for_known_type_strings(
    type_str, strategy_repr
):
    assert repr(get_abi_strategy(type_str)) == strategy_repr



import pytest

from hypothesis import (
    given,
    settings,
)

from faster_eth_abi.packed import (
    is_encodable_packed,
)
from tests.core.common.strategies import (
    single_strs_values,
    tuple_strs_values,
)
from tests.core.common.unit import (
    CORRECT_ENCODINGS,
    NOT_ENCODABLE,
)


@pytest.mark.parametrize(
    "type_str,python_value,_1,_2",
    CORRECT_ENCODINGS,
)
def test_is_encodable_returns_true(type_str, python_value, _1, _2):
    assert is_encodable_packed(type_str, python_value)


@pytest.mark.parametrize(
    "type_str,python_value",
    NOT_ENCODABLE,
)
def test_is_encodable_returns_false(type_str, python_value):
    assert not is_encodable_packed(type_str, python_value)


@settings(deadline=None)
@given(single_strs_values)
def test_is_encodable_returns_true_for_random_valid_values(type_and_value):
    _type, value = type_and_value
    assert is_encodable_packed(_type, value)


@settings(deadline=None)
@given(tuple_strs_values)
def test_is_encodable_returns_true_for_random_valid_tuple_values(type_and_value):
    _type, value = type_and_value
    assert is_encodable_packed(_type, value)



import pytest

from faster_eth_abi import (
    is_encodable_type,
)
from tests.core.common.unit import (
    CORRECT_ENCODINGS,
)


@pytest.mark.parametrize(
    "type_str,_python_value,_1,_2",
    CORRECT_ENCODINGS,
)
def test_is_encodable_type_returns_true(type_str, _python_value, _1, _2):
    assert is_encodable_type(type_str)


def test_is_encodable_type_returns_false():
    assert not is_encodable_type("foo")



"""Base classes for encoder and decoder implementations.

Defines the foundational interface and validation logic for all coder classes.
"""
from typing import (
    Any,
)


class BaseCoder:
    """
    Base class for all encoder and decoder classes.
    """

    is_dynamic = False

    def __init__(self, **kwargs: Any) -> None:
        cls = type(self)

        # Ensure no unrecognized kwargs were given
        for key, value in kwargs.items():
            if not hasattr(cls, key):
                raise AttributeError(
                    "Property {key} not found on {cls_name} class. "
                    "`{cls_name}.__init__` only accepts keyword arguments which are "
                    "present on the {cls_name} class.".format(
                        key=key,
                        cls_name=cls.__name__,
                    )
                )
            setattr(self, key, value)

        # Validate given combination of kwargs
        self.validate()

    def validate(self):
        pass

    @classmethod
    def from_type_str(cls, type_str, registry):  # pragma: no cover
        """
        Used by :any:`ABIRegistry` to get an appropriate encoder or decoder
        instance for the given type string and type registry.
        """
        raise NotImplementedError("Must implement `from_type_str`")



import decimal
from typing import (
    Callable,
    Dict,
    Final,
    Tuple,
)

ABI_DECIMAL_PREC: Final = 999

abi_decimal_context: Final = decimal.Context(prec=ABI_DECIMAL_PREC)
decimal_localcontext: Final = decimal.localcontext

ZERO: Final = decimal.Decimal(0)
TEN: Final = decimal.Decimal(10)

Decimal: Final = decimal.Decimal


def ceil32(x: int) -> int:
    remainder = x % 32
    return x if remainder == 0 else x + 32 - remainder


_unsigned_integer_bounds_cache: Final[Dict[int, Tuple[int, int]]] = {}


def compute_unsigned_integer_bounds(num_bits: int) -> Tuple[int, int]:
    bounds = _unsigned_integer_bounds_cache.get(num_bits)
    if bounds is None:
        bounds = 0, 2**num_bits - 1
        _unsigned_integer_bounds_cache[num_bits] = bounds
    return bounds


_signed_integer_bounds_cache: Final[Dict[int, Tuple[int, int]]] = {}


def compute_signed_integer_bounds(num_bits: int) -> Tuple[int, int]:
    bounds = _signed_integer_bounds_cache.get(num_bits)
    if bounds is None:
        overflow_at = 2 ** (num_bits - 1)
        min_value = -overflow_at
        max_value = overflow_at - 1
        bounds = min_value, max_value
        _signed_integer_bounds_cache[num_bits] = bounds
    return bounds


_unsigned_fixed_bounds_cache: Final[Dict[Tuple[int, int], decimal.Decimal]] = {}


def compute_unsigned_fixed_bounds(
    num_bits: int,
    frac_places: int,
) -> Tuple[decimal.Decimal, decimal.Decimal]:
    upper = _unsigned_fixed_bounds_cache.get((num_bits, frac_places))
    if upper is None:
        int_upper = 2**num_bits - 1

        with decimal_localcontext(abi_decimal_context):
            upper = Decimal(int_upper) * TEN**-frac_places

        _unsigned_fixed_bounds_cache[(num_bits, frac_places)] = upper

    return ZERO, upper


_signed_fixed_bounds_cache: Final[
    Dict[Tuple[int, int], Tuple[decimal.Decimal, decimal.Decimal]]
] = {}


def compute_signed_fixed_bounds(
    num_bits: int,
    frac_places: int,
) -> Tuple[decimal.Decimal, decimal.Decimal]:
    bounds = _signed_fixed_bounds_cache.get((num_bits, frac_places))
    if bounds is None:
        int_lower, int_upper = compute_signed_integer_bounds(num_bits)

        with decimal_localcontext(abi_decimal_context):
            exp = TEN**-frac_places
            lower = Decimal(int_lower) * exp
            upper = Decimal(int_upper) * exp

        bounds = lower, upper
        _signed_fixed_bounds_cache[(num_bits, frac_places)] = bounds

    return bounds


def scale_places(places: int) -> Callable[[decimal.Decimal], decimal.Decimal]:
    """
    Returns a function that shifts the decimal point of decimal values to the
    right by ``places`` places.
    """
    if not isinstance(places, int):
        raise ValueError(
            f"Argument `places` must be int.  Got value {places} "
            f"of type {type(places)}.",
        )

    with decimal_localcontext(abi_decimal_context):
        scaling_factor = TEN**-places

    def f(x: decimal.Decimal) -> decimal.Decimal:
        with decimal_localcontext(abi_decimal_context):
            return x * scaling_factor

    places_repr = f"Eneg{places}" if places > 0 else f"Epos{-places}"
    func_name = f"scale_by_{places_repr}"

    f.__name__ = func_name
    f.__qualname__ = func_name

    return f



"""Internal codec helpers for encoding and decoding sequences of values using the head-tail mechanism.

Provides encode_c and decode_c functions for binary serialization and deserialization of values
according to ABI type specifications.
"""

from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Tuple,
)

from eth_typing import (
    Decodable,
    TypeStr,
)

from faster_eth_abi.utils.validation import (
    validate_bytes_param,
    validate_list_like_param,
)

if TYPE_CHECKING:
    from faster_eth_abi.codec import (
        ABIDecoder,
        ABIEncoder,
    )


def encode_c(
    self: "ABIEncoder",
    types: Iterable[TypeStr],
    args: Iterable[Any],
) -> bytes:
    """
    Encodes the python values in ``args`` as a sequence of binary values of
    the ABI types in ``types`` via the head-tail mechanism.

    :param types: A list or tuple of string representations of the ABI types
        that will be used for encoding e.g.  ``('uint256', 'bytes[]',
        '(int,int)')``
    :param args: A list or tuple of python values to be encoded.

    :returns: The head-tail encoded binary representation of the python
        values in ``args`` as values of the ABI types in ``types``.
    """
    validate_list_like_param(args, "args")

    encoder = self._registry.get_tuple_encoder(*types)

    return encoder.encode(args)


def decode_c(
    self: "ABIDecoder",
    types: Iterable[TypeStr],
    data: Decodable,
    strict: bool = True,
) -> Tuple[Any, ...]:
    """
    Decodes the binary value ``data`` as a sequence of values of the ABI types
    in ``types`` via the head-tail mechanism into a tuple of equivalent python
    values.

    :param types: A list or tuple of string representations of the ABI types that
        will be used for decoding e.g. ``('uint256', 'bytes[]', '(int,int)')``
    :param data: The binary value to be decoded.
    :param strict: If ``False``, dynamic-type decoders will ignore validations such
        as making sure the data is padded to a multiple of 32 bytes or checking that
        padding bytes are zero / empty. ``False`` is how the Solidity ABI decoder
        currently works. However, ``True`` is the default for the faster-eth-abi
        library.

    :returns: A tuple of equivalent python values for the ABI values
        represented in ``data``.
    """
    validate_bytes_param(data, "data")

    decoder = self._registry.get_tuple_decoder(*types, strict=strict)
    stream = self.stream_class(data)

    return decoder(stream)



"""
Script to generate Python typecheck test files for all possible ABI type string argument combinations.

Usage:
    python scripts/generate_overload_tests.py [--impl codec|abi|both] [--suite fixed|var|both] [--sample-1-of-x N] [--seed S] [--lengths L1,L2,...]

This will overwrite test files in:
- tests/typecheck/codec/decode/fixed/len_1/overload_test_data_*.py (tuple-based)
- tests/typecheck/codec/decode/fixed/len_2/overload_test_data_*.py
- tests/typecheck/codec/decode/fixed/len_3/overload_test_data_*.py
- tests/typecheck/codec/decode/variable/len_1/overload_test_data_*.py (iterable-based)
- tests/typecheck/codec/decode/variable/len_2/overload_test_data_*.py
- tests/typecheck/codec/decode/variable/len_3/overload_test_data_*.py
- tests/typecheck/abi/decode/fixed/len_1/overload_test_data_*.py (tuple-based, abi.decode)
- tests/typecheck/abi/decode/variable/len_1/overload_test_data_*.py (iterable-based, abi.decode)
...

Test files are chunked in 10,000-case intervals.

Key features:
- Streams generation: cases are written chunk-by-chunk, not all held in memory.
- Progress bar (tqdm) measures chunk files, not individual cases.
- For TupleXXXIntTypeStr, only 10 representative values are sampled (ending with int8, int16, int64, int128, int256, and their uint variants).
- Each chunk's lines are collected and written in a single operation for efficiency.
- Covers: all int, uint, bytes, bool, string, arrays, and all possible combinations up to length 3, including fallback/unknown types in all permutations.
- If you add new ABI type strings, rerun this script to update the test files.

New options:
- --suite: Restrict to 'fixed', 'var', or 'both' (default: both)
- --sample-1-of-x N: Only generate a random 1/N of the chunk files for each length (default: 1, i.e., all files)
- --seed S: Seed for random sampling (default: 42). Controls which chunk files are selected when sampling, for reproducibility.
- --lengths L1,L2,...: Only generate test data for the specified lengths (e.g., --lengths 1,2,3)
"""

import argparse
import itertools
import math
import pathlib
import random
import re
import sys
from typing import (
    Any,
    Dict,
    Final,
    ForwardRef,
    Iterator,
    List,
    Literal,
    Set,
    Tuple,
    Union,
    get_args,
    get_origin,
)

from tqdm import (
    tqdm,
)
from typing_extensions import (
    LiteralString,
)

MAX_LEN_FIXED: Final = 3
MAX_LEN_VARIABLE: Final = 4
CHUNK_SIZE: Final = 10_000
FILENAME_PREFIX: Final = "overload_test_data_"

ANY_ALIAS: Final = "A"
ASSERT_TYPE_ALIAS: Final = "ck"
BOOL_ALIAS: Final = "bo"
BYTES_ALIAS: Final = "by"
DATA_ALIAS: Final = "D"
FUNC_ALIAS: Final = "f"
HEXADDRESS_ALIAS: Final = "HA"
INT_ALIAS: Final = "i"
STR_ALIAS: Final = "s"
TUPLE_ALIAS: Final = "T"
TUPLE_ANY_ALIAS: Final = "TA"
UNION_ALIAS: Final = "U"


HEADER_CODEC = f"""# AUTOGENERATED: Run scripts/generate_overload_tests.py to update.
# Requires mypy extras
# To run: pip install mypy && mypy tests/typecheck

# alias some things so we can minimize test data size in source control
from typing import Any as {ANY_ALIAS}
from typing import Tuple as {TUPLE_ALIAS}
from typing import Union as {UNION_ALIAS}
from typing_extensions import assert_type as {ASSERT_TYPE_ALIAS}
from eth_typing import HexAddress as {HEXADDRESS_ALIAS}
from faster_eth_abi.codec import ABIDecoder
from faster_eth_abi.registry import ABIRegistry

# more aliases to save a bunch more chars
{BOOL_ALIAS} = bool
{BYTES_ALIAS} = bytes
{INT_ALIAS} = int
{STR_ALIAS} = str

{TUPLE_ANY_ALIAS} = T[A, ...]

THA = T[HA, ...]
Tbo = T[bo, ...]
Tby = T[by, ...]
Ti = T[i, ...]
Ts = T[s, ...]
TUa_bo = T[U[HA, bo], ...]
TUa_by = T[U[HA, by], ...]
TUa_i = T[U[HA, i], ...]
TUa_s = T[U[HA, s], ...]
TUbo_by = T[U[bo, by], ...]
TUbo_i = T[U[bo, i], ...]
TUbo_s = T[U[bo, s], ...]
TUby_i = T[U[by, i], ...]
TUby_s = T[U[by, s], ...]
TUi_s = T[U[i, s], ...]
TUa_bo_by = T[U[bo, by, HA], ...]
TUbo_by_i = T[U[bo, by, i], ...]
TUbo_by_s = T[U[bo, by, s], ...]
TUa_bo_i = T[U[bo, i, HA], ...]
TUa_by_i = T[U[by, i, HA], ...]
TUby_i_s = T[U[by, i, s], ...]

decoder = ABIDecoder(ABIRegistry())

# the callable we are checking
{FUNC_ALIAS} = decoder.decode

# input data
{DATA_ALIAS} = b"\\x00" * 32
"""

HEADER_ABI = f"""# AUTOGENERATED: Run scripts/generate_overload_tests.py to update.
# Requires mypy extras
# To run: pip install mypy && mypy tests/typecheck

# alias some things so we can minimize test data size in source control
from typing import Any as {ANY_ALIAS}
from typing import Tuple as {TUPLE_ALIAS}
from typing import Union as {UNION_ALIAS}
from typing_extensions import assert_type as {ASSERT_TYPE_ALIAS}
from eth_typing import HexAddress as {HEXADDRESS_ALIAS}
from faster_eth_abi import abi

# more aliases to save a bunch more chars
{BOOL_ALIAS} = bool
{BYTES_ALIAS} = bytes
{INT_ALIAS} = int
{STR_ALIAS} = str

{TUPLE_ANY_ALIAS} = T[A, ...]

THA = T[HA, ...]
Tbo = T[bo, ...]
Tby = T[by, ...]
Ti = T[i, ...]
Ts = T[s, ...]
TUa_bo = T[U[HA, bo], ...]
TUa_by = T[U[HA, by], ...]
TUa_i = T[U[HA, i], ...]
TUa_s = T[U[HA, s], ...]
TUbo_by = T[U[bo, by], ...]
TUbo_i = T[U[bo, i], ...]
TUbo_s = T[U[bo, s], ...]
TUby_i = T[U[by, i], ...]
TUby_s = T[U[by, s], ...]
TUi_s = T[U[i, s], ...]
TUa_bo_by = T[U[bo, by, HA], ...]
TUbo_by_i = T[U[bo, by, i], ...]
TUbo_by_s = T[U[bo, by, s], ...]
TUa_bo_i = T[U[bo, i, HA], ...]
TUa_by_i = T[U[by, i, HA], ...]
TUby_i_s = T[U[by, i, s], ...]

# the callable we are checking
{FUNC_ALIAS} = abi.decode

# input data
{DATA_ALIAS} = b"\\x00" * 32
"""

RETURN_TYPE_MAP = {
    "address": HEXADDRESS_ALIAS,
    "address[]": ANY_ALIAS,  # f"{TUPLE_ALIAS}[{HEXADDRESS_ALIAS}, ...]",
    "bytes": BYTES_ALIAS,
    "bytes[]": ANY_ALIAS,  # f"{TUPLE_ALIAS}[{BYTES_ALIAS}, ...]",
    "string": STR_ALIAS,
    "string[]": ANY_ALIAS,  # f"{TUPLE_ALIAS}[{STR_ALIAS}, ...]",
    "int": INT_ALIAS,
    "bool": BOOL_ALIAS,
    "bool[]": ANY_ALIAS,  # f"{TUPLE_ALIAS}[{BOOL_ALIAS}, ...]",
    "?": ANY_ALIAS,
}
for i in range(8, 257, 8):
    RETURN_TYPE_MAP[f"int{i}"] = INT_ALIAS
    RETURN_TYPE_MAP[f"uint{i}"] = INT_ALIAS
    RETURN_TYPE_MAP[f"int{i}[]"] = f"{TUPLE_ALIAS}[{INT_ALIAS}, ...]"
    RETURN_TYPE_MAP[f"uint{i}[]"] = f"{TUPLE_ALIAS}[{INT_ALIAS}, ...]"
for i in range(1, 33):
    RETURN_TYPE_MAP[f"bytes{i}"] = BYTES_ALIAS
    RETURN_TYPE_MAP[f"bytes{i}[]"] = f"{TUPLE_ALIAS}[{BYTES_ALIAS}, ...]"

VAR_TUPLE_ALIASES = {
    # this shouldnt be needed but just in case
    "T[A, ...]": TUPLE_ANY_ALIAS,
    # single type
    "T[HA, ...]": "THA",
    "T[bo, ...]": "Tbo",
    "T[by, ...]": "Tby",
    "T[i, ...]": "Ti",
    "T[s, ...]": "Ts",
    # TODO: add some more later, this is good for now
    "T[U[HA, bo], ...]": "TUa_bo",
    "T[U[HA, by], ...]": "TUa_by",
    "T[U[HA, i], ...]": "TUa_i",
    "T[U[HA, s], ...]": "TUa_s",
    "T[U[bo, by], ...]": "TUbo_by",
    "T[U[bo, i], ...]": "TUbo_i",
    "T[U[bo, s], ...]": "TUbo_s",
    "T[U[by, i], ...]": "TUby_i",
    "T[U[by, s], ...]": "TUby_s",
    "T[U[i, str], ...]": "TUi_s",
    "T[U[bo, by, HA], ...]": "TUa_bo_by",
    "T[U[bo, by, i], ...]": "TUbo_by_i",
    "T[U[bo, by, s], ...]": "TUbo_by_s",
    "T[U[bo, i, HA], ...]": "TUa_bo_i",
    "T[U[by, i, HA], ...]": "TUa_by_i",
    "T[U[by, i, s], ...]": "TUby_i_s",
}


def build_alias_map() -> Dict[str, Any]:
    sys.path.insert(0, str(pathlib.Path(__file__).parent.parent))
    import faster_eth_abi.typing

    alias_map = {}
    for name in dir(faster_eth_abi.typing):
        if not name.startswith("_"):
            obj = getattr(faster_eth_abi.typing, name)
            alias_map[name] = obj
    return alias_map


def all_integers_in_set(s: str, allowed: Set[int]) -> bool:
    """
    Parses the input string for all integer values (including negative numbers)
    and returns True if all found integers are members of the allowed set.
    """
    found = {int(x) for x in re.findall(r"-?\d+", s)}
    return found.issubset(allowed)


def extract_all_literals(typ, alias_map, alias_path=None):
    if alias_path is None:
        alias_path = []
    if typ is str:
        raise ValueError(
            f"Encountered plain 'str' type in alias path: {' -> '.join(alias_path)}."
        )
    origin = get_origin(typ)
    if origin is Union:
        for arg in get_args(typ):
            yield from extract_all_literals(arg, alias_map, alias_path)
        return
    elif origin is Literal:
        if alias_path[-1] in ("BytesArrayTypeStr", "IntArrayTypeStr"):
            # TODO: we need to implement overloads for these, for now we only have the literals defined
            print(f"skipping {alias_path}")
            return
        if alias_path[-1].startswith("Tuple") and alias_path[-1].endswith("IntTypeStr"):
            # for now, we will exclude these. We need to implement them
            # in the actual overloads, which takes a new script. stay tuned.
            return
            for value in get_args(typ):
                if all_integers_in_set(value, {8, 16, 64, 128, 256}):
                    if value not in RETURN_TYPE_MAP:
                        parentheses_stripped = value[1:-1]
                        types = parentheses_stripped.split(",")
                        RETURN_TYPE_MAP[
                            value
                        ] = f"{TUPLE_ALIAS}[{', '.join(RETURN_TYPE_MAP[t] for t in types)}]"
                    yield value
            return
        else:
            yield from get_args(typ)
    elif isinstance(typ, ForwardRef):
        ref_name = typ.__forward_arg__
        if ref_name in alias_map:
            yield from extract_all_literals(
                alias_map[ref_name], alias_map, alias_path + [ref_name]
            )
            return
        raise RuntimeError(
            f"ForwardRef {ref_name!r} not found in alias map (path: {' -> '.join(alias_path)})"
        )
    else:
        raise RuntimeError(
            f"Type {typ!r} is not a Union, Literal, ForwardRef, or str (origin={origin!r}, path: {' -> '.join(alias_path)})"
        )


def expand_literal_values() -> Iterator[LiteralString]:
    alias_map = build_alias_map()
    alias_names = [
        "AddressTypeStr",
        "BytesTypeStr",
        "StringTypeStr",
        "IntegerTypeStr",
        "BoolTypeStr",
        "ArrayTypeStr",
        "TupleTypeStr",
        "DecodesToIntTypeStr",
    ]
    for name in sorted(alias_names):
        if name == "DecodesToIntTypeStr":
            from faster_eth_abi.codec import (
                DecodesToIntTypeStr,
            )

            typ = DecodesToIntTypeStr
        else:
            typ = alias_map[name]
        yield from map(str, extract_all_literals(typ, alias_map, [name]))


ALL_LITERALS: Final = sorted(set(expand_literal_values())) + ["?"]
NUM_LITERALS = len(ALL_LITERALS)
print(f"Found {NUM_LITERALS} ABI type string literals (including fallback).")


def get_expected_type_tuple(types: Tuple[str, ...]) -> str:
    py_types = [ANY_ALIAS if t == "?" else RETURN_TYPE_MAP[t] for t in types]
    return f"{TUPLE_ALIAS}[{', '.join(py_types)}]"


def get_expected_type_iterable(types: Tuple[str, ...]) -> str:
    py_types = [ANY_ALIAS if t == "?" else RETURN_TYPE_MAP[t] for t in types]
    unique_types = sorted(set(py_types))
    if ANY_ALIAS in unique_types:
        expected = TUPLE_ANY_ALIAS
    elif len(unique_types) == 1:
        expected = f"{TUPLE_ALIAS}[{unique_types[0]}, ...]"
    else:
        expected = f"{TUPLE_ALIAS}[{UNION_ALIAS}[{', '.join(unique_types)}], ...]"
    return VAR_TUPLE_ALIASES.get(expected) or expected


def compute_total_chunks(L: int) -> int:
    total_cases = NUM_LITERALS**L
    return math.ceil(total_cases / CHUNK_SIZE)


def stream_cases_and_write_files(
    mode: Literal["fixed", "variable"],
    impl: Literal["abi", "codec"],
    sample_1_of_x: int = 1,
    seed: int = 42,
    lengths: List[int] = None,
) -> int:
    FOLDER = "fixed" if mode == "fixed" else "variable"
    if lengths is None:
        MAX_LEN = MAX_LEN_FIXED if mode == "fixed" else MAX_LEN_VARIABLE
        lengths = list(range(1, MAX_LEN + 1))

    case_counter = 0

    for L in lengths:
        total_chunks = compute_total_chunks(L)
        chunk_indices = list(range(1, total_chunks + 1))
        if sample_1_of_x > 1:
            random.seed(seed)
            num_to_pick = max(1, total_chunks // sample_1_of_x)
            chunk_indices = sorted(random.sample(chunk_indices, num_to_pick))

        with tqdm(
            total=len(chunk_indices),
            desc=f"Streaming {mode} chunks for {impl} (len_{L})",
        ) as progress:
            PATH_WITHOUT_NUMBER = (
                f"tests/typecheck/{impl}/decode/{FOLDER}/len_{L}/{FILENAME_PREFIX}"
            )
            chunk_idx = 1
            chunk_lines = []
            file_counter = 0

            def format_case(typestr: str, expected: str) -> str:
                nonlocal case_counter, file_counter
                case = f"{ASSERT_TYPE_ALIAS}({FUNC_ALIAS}({typestr}, {DATA_ALIAS}), {expected})  # {mode} case {case_counter}\n"
                case_counter += 1
                file_counter += 1
                return case

            if mode == "fixed":

                def addline(case: Tuple[str, ...]) -> Iterator[str]:
                    typestr = (
                        f"('{case[0]}',)"
                        if len(case) == 1
                        else f"({', '.join(map(repr, case))})"
                    )
                    expected = get_expected_type_tuple(case)
                    chunk_lines.append(format_case(typestr, expected))
                    # chunk_lines.append(format_case("ABICodec.decode", f"codec, {typestr}", expected)
                    # chunk_lines.append(format_case("ABIDecoder.decode", f"codec, {typestr}", expected)
                    # chunk_lines.append(format_case("ABIDecoder.decode", f"decoder, {typestr}", expected)

            else:

                def addline(case: Tuple[str, ...]) -> Iterator[str]:
                    typestr = f"[{', '.join(map(repr, case))}]"
                    expected = get_expected_type_iterable(case)
                    chunk_lines.append(format_case(typestr, expected))
                    # chunk_lines.append(format_case("ABICodec.decode", f"codec, {typestr}", expected)
                    # chunk_lines.append(format_case("ABIDecoder.decode", f"codec, {typestr}", expected)
                    # chunk_lines.append(format_case("ABIDecoder.decode", f"decoder, {typestr}", expected)

            def write_file() -> None:
                nonlocal chunk_idx
                if sample_1_of_x == 1 or chunk_idx in chunk_indices:
                    path = pathlib.Path(f"{PATH_WITHOUT_NUMBER}{chunk_idx:05d}.py")
                    path.parent.mkdir(parents=True, exist_ok=True)
                    with open(path, "w") as f:
                        f.write(HEADER_ABI if impl == "abi" else HEADER_CODEC)
                        f.writelines(chunk_lines)
                    progress.update(1)
                chunk_idx += 1
                chunk_lines.clear()

            for combo in itertools.product(ALL_LITERALS, repeat=L):
                if file_counter and file_counter % CHUNK_SIZE == 0:
                    write_file()
                addline(combo)

            if chunk_lines:
                write_file()

            print(f"finished generating cases for length {L}")

    return case_counter


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--impl", choices=["codec", "abi", "both"], default="both")
    parser.add_argument(
        "--suite",
        choices=["fixed", "variable", "both"],
        default="both",
        help="Restrict to 'fixed', 'var', or 'both' test suites",
    )
    parser.add_argument(
        "--sample-1-of-x",
        type=int,
        default=1,
        help="Only generate a random 1/N of the chunk files for each length (default: 1, i.e., all files)",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=42,
        help="Seed for random sampling (default: 42). Controls which chunk files are selected when sampling, for reproducibility.",
    )
    parser.add_argument(
        "--lengths",
        type=str,
        help="Comma-separated list of lengths to generate (e.g., 1,2,3,4)",
    )
    args = parser.parse_args()

    impls = ["codec", "abi"] if args.impl == "both" else [args.impl]
    modes = ["fixed", "variable"] if args.suite == "both" else [args.suite]
    lengths = None
    if args.lengths:
        lengths = [int(x) for x in args.lengths.split(",") if x.strip()]
    for impl in impls:
        for mode in modes:
            print(
                f"Streaming {'tuple' if mode == 'fixed' else 'iterable'}-based test cases for {impl}..."
            )
            stream_cases_and_write_files(
                mode=mode,
                impl=impl,
                sample_1_of_x=args.sample_1_of_x,
                seed=args.seed,
                lengths=lengths,
            )

    print("Test generation complete.")


if __name__ == "__main__":
    main()



import pytest

from hypothesis import (
    example,
    given,
    strategies as st,
)

from faster_eth_abi.exceptions import (
    ABITypeError,
    ParseError,
)
from faster_eth_abi.grammar import (
    TYPE_ALIASES,
    BasicType,
    TupleType,
    normalize,
    parse,
)
from tests.core.common.strategies import (
    malformed_type_strs,
    type_strs,
)


@pytest.mark.parametrize(
    "type_str, expected_type",
    (
        ("custom", BasicType("custom")),
        ("Custom", BasicType("Custom")),
        ("CUSTOM", BasicType("CUSTOM")),
        ("Custom[2]", BasicType("Custom", None, ((2,),))),
        ("uint", BasicType("uint")),
        ("uint256", BasicType("uint", 256)),
        ("uint256[]", BasicType("uint", 256, ((),))),
        ("uint256[][1][23][]", BasicType("uint", 256, ((), (1,), (23,), ()))),
        ("uint[]", BasicType("uint", None, ((),))),
        ("int[2]", BasicType("int", None, ((2,),))),
        ("fixed128x19", BasicType("fixed", (128, 19))),
        ("ufixed128x19[][]", BasicType("ufixed", (128, 19), ((), ()))),
        (
            "(int,ufixed128x19[][])",
            TupleType(
                (
                    BasicType("int"),
                    BasicType("ufixed", (128, 19), ((), ())),
                )
            ),
        ),
        (
            "(int,(address,function))",
            TupleType(
                (
                    BasicType("int"),
                    TupleType(
                        (
                            BasicType("address"),
                            BasicType("function"),
                        )
                    ),
                )
            ),
        ),
        (
            "((address,(fixed,bool)),(uint256[][1][],fixed128x19))",
            TupleType(
                (
                    TupleType(
                        (
                            BasicType("address"),
                            TupleType(
                                (
                                    BasicType("fixed"),
                                    BasicType("bool"),
                                )
                            ),
                        )
                    ),
                    TupleType(
                        (
                            BasicType("uint", 256, ((), (1,), ())),
                            BasicType("fixed", (128, 19)),
                        )
                    ),
                )
            ),
        ),
        ("()", TupleType(tuple())),
        (
            "(int,(),())",
            TupleType((BasicType("int"), TupleType(tuple()), TupleType(tuple()))),
        ),
    ),
)
def test_parsing_with_parsimonious_grammar_and_node_visitor_works(
    type_str, expected_type
):
    assert parse(type_str) == expected_type


@pytest.mark.parametrize(
    "type_str, expected_type",
    (
        ("uint", BasicType("uint", 256)),
        ("uint256[]", BasicType("uint", 256, ((),))),
        ("function", BasicType("bytes", 24)),
        ("fixed", BasicType("fixed", (128, 18))),
        ("ufixed", BasicType("ufixed", (128, 18))),
    ),
)
def test_normalizing_and_parsing_works(type_str, expected_type):
    assert parse(normalize(type_str)) == expected_type


@given(malformed_type_strs, st.none())
@example("int0", 4)
@example("fixed128x0", 9)
@example("ufixed128x0", 10)
@example("int!", 4)
@example("(int, bool)", 6)
@example("(int,bool,)", 11)
@example("(int,(address,uint256[][])", 27)
def test_parsing_invalid_type_str_causes_parse_error(type_str, error_col):
    if error_col is not None:
        pattern = rf"Parse error at .* \(column {error_col}\)"
    else:
        pattern = r"Parse error at .*"

    with pytest.raises(ParseError, match=pattern):
        parse(type_str)


def test_parse_raises_type_error_for_wrong_data_type():
    with pytest.raises(TypeError):
        parse(b"uint256")


@given(type_strs)
def test_end_to_end_parsing_and_collapsing(type_str):
    assert parse(type_str).to_type_str() == type_str


@pytest.mark.parametrize(
    "type_str, pattern",
    (
        ("string1x1", "string type cannot have suffix"),
        ("string33", "string type cannot have suffix"),
        ("bytes1x1", "no suffix or a numerical"),
        ("bytes33", "maximum 32 bytes"),
        ("int", "must have numerical suffix"),
        ("int7", "size out of bounds"),
        ("int257", "size out of bounds"),
        ("int9", "must be multiple of 8"),
        ("uint", "must have numerical suffix"),
        ("uint7", "size out of bounds"),
        ("uint257", "size out of bounds"),
        ("uint9", "must be multiple of 8"),
        ("fixed", "must have suffix of form"),
        ("fixed8", "must have suffix of form"),
        ("fixed7x19", "fixed size out of bounds"),
        ("fixed257x19", "fixed size out of bounds"),
        ("fixed17x19", "must be multiple of 8"),
        ("fixed128x81", "exponent size out of bounds"),
        ("ufixed", "must have suffix of form"),
        ("ufixed8", "must have suffix of form"),
        ("ufixed7x19", "fixed size out of bounds"),
        ("ufixed257x19", "fixed size out of bounds"),
        ("ufixed17x19", "must be multiple of 8"),
        ("ufixed128x81", "exponent size out of bounds"),
        ("hash", "must have numerical suffix"),
        ("hash128x128", "must have numerical suffix"),
        ("address128", "cannot have suffix"),
        ("address128x128", "cannot have suffix"),
        (
            "(int256,(address,(ufixed7x19,fixed)))",
            "For 'ufixed7x19' type at column 19 "
            r"in '\(int256,\(address,\(ufixed7x19,fixed\)\)\)': "
            "fixed size out of bounds",
        ),
        (
            "(int256,(address,(ufixed128x19,fixed)))",
            "For 'fixed' type at column 32 "
            r"in '\(int256,\(address,\(ufixed128x19,fixed\)\)\)': "
            "fixed type must have suffix of form",
        ),
    ),
)
def test_invalid_abi_types(type_str, pattern):
    with pytest.raises(ABITypeError, match=pattern):
        parse(type_str).validate()


@pytest.mark.parametrize(
    "type_str",
    (
        "string",
        "bytes24",
        "int256",
        "uint128",
        "fixed128x19",
        "ufixed256x19",
        "hash256",
        "address",
        "(int256,(address,(ufixed128x19,fixed128x19)))",
    ),
)
def test_valid_abi_types(type_str):
    parse(type_str).validate()


@pytest.mark.parametrize(
    "type_str, normalized",
    tuple(TYPE_ALIASES.items())
    + (
        ("(int,uint,fixed,ufixed)", "(int256,uint256,fixed128x18,ufixed128x18)"),
        ("(function,function,(function,byte))", "(bytes24,bytes24,(bytes24,bytes1))"),
    ),
)
def test_normalize(type_str, normalized):
    assert normalize(type_str) == normalized


def test_basic_type_item_type_throws_errors():
    bt = parse("int256")

    pattern = "Cannot determine item type for non-array type 'int256'"
    with pytest.raises(ValueError, match=pattern):
        bt.item_type


@pytest.mark.parametrize(
    "type_str",
    (
        "string",
        "bytes",
        "Custom[]",
        "(bool,Custom[])[2]",
        "int[]",
        "int[][2]",
        "((int[][2],bool),int)",
        "(bool,bool)[][2]",
    ),
)
def test_abi_type_dynamic_types(type_str):
    abi_type = parse(type_str)
    assert abi_type.is_dynamic


@pytest.mark.parametrize(
    "type_str",
    (
        "bytes2",
        "Custom",
        "int[2]",
        "int[2][2]",
        "((int[2][2],bool),int)",
        "(bool,bool)[1][2]",
    ),
)
def test_abi_type_static_types(type_str):
    abi_type = parse(type_str)
    assert not abi_type.is_dynamic


@pytest.mark.parametrize(
    "type_str",
    (
        "bytes[][2]",
        "bytes[]",
        "(bool,bool)[][2]",
        "(bool,bool)[]",
    ),
)
def test_abi_type_has_dynamic_arrlist(type_str):
    abi_type = parse(type_str)
    assert abi_type._has_dynamic_arrlist


@pytest.mark.parametrize(
    "type_str",
    (
        "bytes[2][2]",
        "bytes",
        "string",
        "(int[],bool)",
    ),
)
def test_abi_type_lacks_dynamic_arrlist(type_str):
    abi_type = parse(type_str)
    assert not abi_type._has_dynamic_arrlist


@pytest.mark.parametrize(
    "type_str",
    (
        "bytes[2]",
        "Custom[2]",
        "bytes[][]",
        "(int[],bool)[]",
        "(int[],bool)[][2]",
    ),
)
def test_abi_type_is_array(type_str):
    abi_type = parse(type_str)
    assert abi_type.is_array


@pytest.mark.parametrize(
    "type_str",
    (
        "Custom",
        "bytes",
        "bytes2",
        "(int[],bool)",
    ),
)
def test_abi_type_is_not_array(type_str):
    abi_type = parse(type_str)
    assert not abi_type.is_array



import pytest

from faster_eth_abi.base import (
    BaseCoder,
)
from faster_eth_abi.from_type_str import (
    parse_tuple_type_str,
    parse_type_str,
)
from faster_eth_abi.grammar import (
    BasicType,
)


class Coder(BaseCoder):
    base = None
    sub = None

    def validate(self):
        if self.base is None:
            raise ValueError("Setting missing")

    @parse_type_str(with_arrlist=False)
    def from_type_str(cls, abi_type, registry):
        return cls(base=abi_type.base, sub=abi_type.sub)


class ArrayCoder(Coder):
    arrlist = None

    @parse_type_str(with_arrlist=True)
    def from_type_str(cls, abi_type, registry):
        return cls(base=abi_type.base, sub=abi_type.sub, arrlist=abi_type.arrlist)


class IntCoder(Coder):
    @parse_type_str("int", with_arrlist=False)
    def from_type_str(cls, abi_type, registry):
        return cls(base=abi_type.base, sub=abi_type.sub)


class IntArrayCoder(ArrayCoder):
    @parse_type_str("int", with_arrlist=True)
    def from_type_str(cls, abi_type, registry):
        return cls(base=abi_type.base, sub=abi_type.sub, arrlist=abi_type.arrlist)


class TupleCoder(BaseCoder):
    components = None

    @parse_tuple_type_str
    def from_type_str(cls, abi_type, registry):
        return cls(components=abi_type.components)


def test_base_coder_requires_that_settings_are_known():
    pattern = "only accepts keyword arguments which are present"
    with pytest.raises(AttributeError, match=pattern):
        Coder(foo="bar")


def test_base_coder_validates_settings():
    with pytest.raises(ValueError, match="Setting missing"):
        Coder(sub="foo")


def test_base_coder_assigns_all_kwargs_to_instance():
    coder = Coder(base="foo", sub="bar")

    assert coder.base == "foo"
    assert coder.sub == "bar"


def test_parse_type_str_normalizes_and_parses_types():
    coder1 = Coder.from_type_str("int", None)

    assert coder1.base == "int"
    assert coder1.sub == 256


def test_parse_type_str_validates_standard_types():
    with pytest.raises(ValueError, match="integer size must be multiple of 8"):
        Coder.from_type_str("int255", None)


def test_parse_type_str_requires_certain_base_type_if_given():
    with pytest.raises(ValueError, match="expected type with base 'int'"):
        IntCoder.from_type_str("uint256", None)

    with pytest.raises(ValueError, match="expected type with base 'int'"):
        IntArrayCoder.from_type_str("uint256[]", None)


def test_parse_type_str_doesnt_require_base_type_if_not_given():
    coder1 = Coder.from_type_str("uint256", None)
    assert coder1.base == "uint"
    assert coder1.sub == 256

    coder2 = ArrayCoder.from_type_str("uint256[]", None)
    assert coder2.base == "uint"
    assert coder2.sub == 256
    assert coder2.arrlist == ((),)


def test_parse_type_str_distinguishes_non_array_and_array_types():
    with pytest.raises(ValueError, match="expected type with no array dimension list"):
        Coder.from_type_str("int256[]", None)
    with pytest.raises(ValueError, match="expected type with array dimension list"):
        ArrayCoder.from_type_str("int256", None)

    with pytest.raises(ValueError, match="expected type with no array dimension list"):
        IntCoder.from_type_str("int256[]", None)
    with pytest.raises(ValueError, match="expected type with array dimension list"):
        IntArrayCoder.from_type_str("int256", None)


def test_parse_tuple_type_str_normalizes_and_parses_types():
    tuple_coder = TupleCoder.from_type_str("(int,int)", None)

    assert tuple_coder.components == (
        BasicType("int", 256, None),
        BasicType("int", 256, None),
    )


def test_parse_tuple_type_str_validates_standard_types():
    with pytest.raises(ValueError, match="integer size must be multiple of 8"):
        TupleCoder.from_type_str("(int255,int)", None)


def test_parse_tuple_type_str_distinguishes_non_tuple_and_tuple_types():
    with pytest.raises(ValueError, match="Cannot create TupleCoder for non-tuple type"):
        TupleCoder.from_type_str("int256", None)



import pytest
import copy
import functools
import itertools
import operator

from faster_eth_abi.registry import (
    Equals,
    PredicateMapping,
)


@pytest.fixture
def mapping():
    return PredicateMapping("registry")


equals_foo = Equals("foo")
equals_bar = Equals("bar")


def test_can_add_predicates_and_find_corresponding_values(mapping: PredicateMapping):
    mapping.add(equals_foo, "foo value")
    mapping.add(equals_bar, "bar value")

    assert mapping.find("foo") == "foo value"
    assert mapping.find("bar") == "bar value"


def test_cannot_add_same_predicate_twice(mapping: PredicateMapping):
    mapping.add(equals_foo, "foo value")

    pattern = r"Matcher .* already exists in registry"
    with pytest.raises(ValueError, match=pattern):
        mapping.add(equals_foo, "foo value")


def test_cannot_add_two_predicates_with_same_label(mapping: PredicateMapping):
    mapping.add(equals_foo, "foo value", "some label")

    pattern = r"Matcher .* with label 'some label' already exists in registry"
    with pytest.raises(ValueError, match=pattern):
        mapping.add(equals_bar, "bar value", "some label")


def test_no_matching_predicates_should_raise_error(mapping: PredicateMapping):
    mapping.add(equals_foo, "foo value")
    mapping.add(equals_bar, "bar value")

    with pytest.raises(ValueError, match=r"No matching entries for 'baz' in registry"):
        mapping.find("baz")


def test_multiple_matching_predicates_should_raise_error(mapping: PredicateMapping):
    mapping.add(Equals("baz"), "baz value")
    mapping.add(functools.partial(operator.eq, "baz"), "baz value")

    with pytest.raises(ValueError) as e:
        mapping.find("baz")

    e.match(r"Multiple matching entries for 'baz' in registry")
    e.match(r"<Equals \(== 'baz'\)>")
    e.match(r"functools\.partial")


def test_can_remove_predicates_by_reference(mapping: PredicateMapping):
    mapping.add(equals_foo, "foo value")
    assert mapping.find("foo") == "foo value"

    mapping.remove(equals_foo)
    with pytest.raises(ValueError, match=r"No matching"):
        mapping.find("foo")


def test_removing_non_existent_predicate_raises_error(mapping: PredicateMapping):
    with pytest.raises(KeyError, match=r"Matcher .* not found in registry"):
        mapping.remove(equals_foo)


def test_removing_labeled_predicate_by_reference_also_removes_label(
    mapping: PredicateMapping,
):
    mapping.add(equals_foo, "foo value", "foo label")
    mapping.remove(equals_foo)

    # Just a bit of white-box testing here
    assert len(mapping._values) == 0
    assert len(mapping._labeled_predicates) == 0


def test_can_remove_predicates_by_label_if_label_was_given(mapping: PredicateMapping):
    mapping.add(equals_foo, "foo value", "foo label")
    assert mapping.find("foo") == "foo value"

    mapping.remove("foo label")
    with pytest.raises(ValueError, match=r"No matching"):
        mapping.find("foo")


def test_removing_non_existent_label_raises_error(mapping: PredicateMapping):
    with pytest.raises(KeyError, match=r"Label 'asdf' not found in registry"):
        mapping.remove("asdf")


def test_removing_unsupported_type_raises_error(mapping: PredicateMapping):
    with pytest.raises(
        TypeError, match=r"must be callable or string.*got <class \'bool\'>"
    ):
        mapping.remove(True)


def test_copying_copies_internal_mappings_but_not_predicates_or_values(
    mapping: PredicateMapping,
):
    mapping.add(equals_foo, "foo value", "foo label")

    c1 = mapping.copy()
    c2 = copy.copy(mapping)
    c3 = copy.deepcopy(mapping)

    for x, y in itertools.combinations((mapping, c1, c2, c3), 2):
        # Objects are different
        assert id(x) != id(y)

        # Names are the same
        assert x._name == y._name

        # Internal mapping objects are different
        assert id(x._values) != id(y._values)
        assert id(x._labeled_predicates) != id(y._labeled_predicates)

        # Values for 'foo' can be found in both mappings
        assert x.find("foo") == "foo value"
        assert y.find("foo") == "foo value"



import abc
import decimal
from _typeshed import Incomplete
from eth_typing import HexAddress
from faster_eth_abi._decoding import (
    decode_dynamic_array as decode_dynamic_array,
    decode_head_tail as decode_head_tail,
    decode_sized_array as decode_sized_array,
    decode_tuple as decode_tuple,
    decoder_fn_boolean as decoder_fn_boolean,
    get_value_byte_size as get_value_byte_size,
    read_fixed_byte_size_data_from_stream as read_fixed_byte_size_data_from_stream,
    split_data_and_padding_fixed_byte_size as split_data_and_padding_fixed_byte_size,
    validate_padding_bytes_fixed_byte_size as validate_padding_bytes_fixed_byte_size,
    validate_padding_bytes_signed_integer as validate_padding_bytes_signed_integer,
    validate_pointers_array as validate_pointers_array,
)
from faster_eth_abi.base import BaseCoder as BaseCoder
from faster_eth_abi.exceptions import (
    InsufficientDataBytes as InsufficientDataBytes,
    NonEmptyPaddingBytes as NonEmptyPaddingBytes,
)
from faster_eth_abi.from_type_str import (
    parse_tuple_type_str as parse_tuple_type_str,
    parse_type_str as parse_type_str,
)
from faster_eth_abi.io import ContextFramesBytesIO as ContextFramesBytesIO
from faster_eth_abi.typing import T as T
from faster_eth_abi.utils.numeric import (
    TEN as TEN,
    abi_decimal_context as abi_decimal_context,
    ceil32 as ceil32,
)
from functools import cached_property as cached_property
from typing import Any, Callable, Final, Generic, TypeVar, final

TByteStr = TypeVar("TByteStr", bytes, str)

class BaseDecoder(BaseCoder, Generic[T], metaclass=abc.ABCMeta):
    """
    Base class for all decoder classes.  Subclass this if you want to define a
    custom decoder class.  Subclasses must also implement
    :any:`BaseCoder.from_type_str`.
    """

    strict: bool
    @abc.abstractmethod
    def decode(self, stream: ContextFramesBytesIO) -> T:
        """
        Decodes the given stream of bytes into a python value.  Should raise
        :any:`exceptions.DecodingError` if a python value cannot be decoded
        from the given byte stream.
        """

    def __call__(self, stream: ContextFramesBytesIO) -> T: ...

class HeadTailDecoder(BaseDecoder[T]):
    """
    Decoder for a dynamic element of a dynamic container (a dynamic array, or a sized
    array or tuple that contains dynamic elements). A dynamic element consists of a
    pointer, aka offset, which is located in the head section of the encoded container,
    and the actual value, which is located in the tail section of the encoding.
    """

    is_dynamic: bool
    tail_decoder: Final[Incomplete]
    def __init__(
        self,
        tail_decoder: (
            HeadTailDecoder[T]
            | SizedArrayDecoder[T]
            | DynamicArrayDecoder[T]
            | ByteStringDecoder[T]
        ),
        **kwargs: Any
    ) -> None: ...
    def decode(self, stream: ContextFramesBytesIO) -> T: ...
    __call__ = decode

class TupleDecoder(BaseDecoder[tuple[T, ...]]):
    decoders: tuple[BaseDecoder[T], ...]
    is_dynamic: Incomplete
    len_of_head: Incomplete
    def __init__(self, decoders: tuple[BaseDecoder[T], ...], **kwargs: Any) -> None: ...
    def validate(self) -> None: ...
    @final
    def validate_pointers(self, stream: ContextFramesBytesIO) -> None: ...
    def decode(self, stream: ContextFramesBytesIO) -> tuple[T, ...]: ...
    __call__ = decode
    @parse_tuple_type_str
    def from_type_str(cls, abi_type, registry): ...

class SingleDecoder(BaseDecoder[T]):
    decoder_fn: Incomplete
    def validate(self) -> None: ...
    def validate_padding_bytes(self, value: Any, padding_bytes: bytes) -> None: ...
    def decode(self, stream: ContextFramesBytesIO) -> T: ...
    __call__ = decode
    def read_data_from_stream(self, stream: ContextFramesBytesIO) -> bytes: ...
    def split_data_and_padding(self, raw_data: bytes) -> tuple[bytes, bytes]: ...

class BaseArrayDecoder(BaseDecoder[tuple[T, ...]]):
    item_decoder: BaseDecoder
    def __init__(self, **kwargs: Any) -> None: ...
    def decode(self, stream: ContextFramesBytesIO) -> tuple[T, ...]: ...
    def validate(self) -> None: ...
    def from_type_str(cls, abi_type, registry): ...
    def validate_pointers(self, stream: ContextFramesBytesIO, array_size: int) -> None:
        """
        Verify that all pointers point to a valid location in the stream.
        """

class SizedArrayDecoder(BaseArrayDecoder[T]):
    array_size: int
    is_dynamic: Incomplete
    def __init__(self, **kwargs) -> None: ...
    def decode(self, stream: ContextFramesBytesIO) -> tuple[T, ...]: ...
    __call__ = decode

class DynamicArrayDecoder(BaseArrayDecoder[T]):
    is_dynamic: bool
    def decode(self, stream: ContextFramesBytesIO) -> tuple[T, ...]: ...
    __call__ = decode

class FixedByteSizeDecoder(SingleDecoder[T]):
    decoder_fn: Callable[[bytes], T]
    value_bit_size: int
    data_byte_size: int
    is_big_endian: bool
    validate_padding_bytes: Incomplete
    def __init__(self, **kwargs) -> None: ...
    def validate(self) -> None: ...
    def read_data_from_stream(self, stream: ContextFramesBytesIO) -> bytes: ...
    def split_data_and_padding(self, raw_data: bytes) -> tuple[bytes, bytes]: ...

class Fixed32ByteSizeDecoder(FixedByteSizeDecoder[T]):
    data_byte_size: int

class BooleanDecoder(Fixed32ByteSizeDecoder[bool]):
    value_bit_size: int
    is_big_endian: bool
    decoder_fn: Incomplete
    def from_type_str(cls, abi_type, registry): ...

class AddressDecoder(Fixed32ByteSizeDecoder[HexAddress]):
    value_bit_size: Incomplete
    is_big_endian: bool
    decoder_fn: Incomplete
    def from_type_str(cls, abi_type, registry): ...

class UnsignedIntegerDecoder(Fixed32ByteSizeDecoder[int]):
    decoder_fn: staticmethod[[bytes], int]
    is_big_endian: bool
    def from_type_str(cls, abi_type, registry): ...

decode_uint_256: Incomplete

class UnsignedIntegerDecoderCached(UnsignedIntegerDecoder):
    decoder_fn: Callable[[bytes], int]
    maxsize: Final[int | None]
    def __init__(self, maxsize: int | None = None, **kwargs: Any) -> None: ...

class SignedIntegerDecoder(Fixed32ByteSizeDecoder[int]):
    is_big_endian: bool
    def __init__(self, **kwargs: Any) -> None: ...
    @cached_property
    def neg_threshold(self) -> int: ...
    @cached_property
    def neg_offset(self) -> int: ...
    def decoder_fn(self, data: bytes) -> int: ...
    def validate_padding_bytes(self, value: Any, padding_bytes: bytes) -> None: ...
    def from_type_str(cls, abi_type, registry): ...

class SignedIntegerDecoderCached(SignedIntegerDecoder):
    decoder_fn: Callable[[bytes], int]
    maxsize: Final[int | None]
    def __init__(self, maxsize: int | None = None, **kwargs: Any) -> None: ...

class BytesDecoder(Fixed32ByteSizeDecoder[bytes]):
    is_big_endian: bool
    @staticmethod
    def decoder_fn(data: bytes) -> bytes: ...
    def from_type_str(cls, abi_type, registry): ...

class BaseFixedDecoder(Fixed32ByteSizeDecoder[decimal.Decimal]):
    frac_places: int
    is_big_endian: bool
    @cached_property
    def denominator(self) -> decimal.Decimal: ...
    def validate(self) -> None: ...

class UnsignedFixedDecoder(BaseFixedDecoder):
    def decoder_fn(self, data: bytes) -> decimal.Decimal: ...
    def from_type_str(cls, abi_type, registry): ...

class SignedFixedDecoder(BaseFixedDecoder):
    @cached_property
    def neg_threshold(self) -> int: ...
    @cached_property
    def neg_offset(self) -> int: ...
    @cached_property
    def expected_padding_pos(self) -> bytes: ...
    @cached_property
    def expected_padding_neg(self) -> bytes: ...
    def decoder_fn(self, data: bytes) -> decimal.Decimal: ...
    def validate_padding_bytes(self, value: Any, padding_bytes: bytes) -> None: ...
    def from_type_str(cls, abi_type, registry): ...

class ByteStringDecoder(SingleDecoder[TByteStr]):
    is_dynamic: bool
    @staticmethod
    def decoder_fn(data: bytes) -> bytes: ...
    def read_data_from_stream(self, stream: ContextFramesBytesIO) -> bytes: ...
    def validate_padding_bytes(self, value: Any, padding_bytes: bytes) -> None: ...
    def from_type_str(cls, abi_type, registry): ...

class StringDecoder(ByteStringDecoder[str]):
    bytes_errors: Final[Incomplete]
    def __init__(self, handle_string_errors: str = "strict") -> None: ...
    def from_type_str(cls, abi_type, registry): ...
    def decode(self, stream: ContextFramesBytesIO) -> str: ...
    __call__ = decode
    @staticmethod
    def decoder_fn(data: bytes, handle_string_errors: str = "strict") -> str: ...
